from pathlib import Path
from summawise import ai, utils
from summawise.files.metadata import FileMetadata
from summawise.files import utils as FileUtils
from summawise.settings import Settings


def process_dir(dir_path: Path, delete: bool = True) -> ai.Resources:
    _ = delete

    resources: ai.Resources
    files = FileUtils.list_files(dir_path)
    # TODO(justin): look into improving/validating approach
    filtered = FileUtils.filter_files(files)
    try:
        print(
            f"Directory scan located and validated {filtered.valid_count}/{filtered.total_count} files.")
        resources = ai.create_vector_store(dir_path.name, filtered.files)
        print(f"Vector store created with ID: {resources.vector_store_id}")
    except Exception as ex:
        raise Exception(f"Error creating vector store [{type(ex)}]: {ex}")

    return resources


def process_file(file_path: Path, delete: bool = False) -> ai.Resources:
    # TODO(justin):
    # - add archive support (.zip, .tar.gz) - extract, call process_dir
    # - maybe add automatic extraction of text from pdf or html (undecided)

    # use settings class (singleton)
    settings = Settings()  # type: ignore

    # file extension to cache file metadata (json or bin)
    ext = settings.data_mode.ext()

    metadata = FileMetadata.create_from_path(file_path)
    hash = metadata.hash
    hash_path = utils.fp(utils.get_summawise_dir() / "files" / f"{hash}.{ext}")

    if not hash_path.exists():
        try:
            resources = ai.create_vector_store(file_path.stem, [file_path])
            metadata.vector_store_id = resources.vector_store_id
            metadata.file_id = next(iter(resources.file_ids))
            metadata.save_to_file(
                file_path=hash_path,
                mode=settings.data_mode,
                compress=settings.compression,
                pretty_json=True
            )
            print(f"Vector store created with ID: {metadata.vector_store_id}")
        except Exception as ex:
            raise Exception(f"Error creating vector store [{type(ex)}]: {ex}")
    else:
        metadata = FileMetadata.from_file(hash_path, settings.data_mode)
        print(
            f"Restored vector store ID from cache: {metadata.vector_store_id}")

    if delete and file_path.exists():
        file_path.unlink()

    return ai.Resources(
        vector_store_ids=[metadata.vector_store_id],
        file_ids=[metadata.file_id]
    )



import json
import textwrap
from typing_extensions import override
from typing import List, Optional, NamedTuple, Dict, Set
from pathlib import Path
from dataclasses import dataclass, field
from openai import OpenAI, AssistantEventHandler
from openai.types.file_object import FileObject
from openai.types.beta import Thread, Assistant, VectorStore
from openai.types.beta.threads import TextContentBlock, TextDelta, Message, Text
from openai.types.beta.threads.runs import ToolCall, ToolCallDelta
from openai.types.beta.thread_create_params import ToolResources, Message as TCPMessage
from prompt_toolkit import ANSI, HTML, print_formatted_text as print
from pygments.formatters import Terminal256Formatter
from summawise.files.cache import FileCacheObj
from summawise.files import utils as FileUtils
from summawise.settings import Settings
from summawise import utils

Client: OpenAI
FileCache: FileCacheObj


@dataclass
class Resources:
    vector_store_ids: List[str] = field(default_factory=list)
    file_ids: List[str] = field(default_factory=list)
    file_contents: Dict[Path, str] = field(default_factory=dict)

    @property
    def vector_store_id(self):
        if not len(self.vector_store_ids):
            raise ValueError("Resources object has no vector store id.")
        elif len(self.vector_store_ids) > 1:
            raise ValueError(
                "Resources object has more than 1 vector store id associated with it.")
        return self.vector_store_ids[0]


class EventHandler(AssistantEventHandler):

    auto_print: bool
    response_text: str

    _tool_calls: Dict[str, ToolCall]
    _completed_tool_calls: Set[str]

    def __init__(self, auto_print: bool = False):
        self.auto_print = auto_print
        self.response_text = ""

        self._tool_calls = dict()
        self._completed_tool_calls = set()

        super().__init__()

    @override
    def on_text_created(self, text: Text) -> None:
        _ = text
        if self.auto_print:
            print("\nsummawise > ", end="", flush=True)

    @override
    def on_text_delta(self, delta: TextDelta, snapshot: Text) -> None:
        _ = snapshot
        if self.auto_print:
            print(delta.value, end="", flush=True)

    @override
    def on_message_done(self, message: Message) -> None:
        print(flush=True)  # new line
        for content in message.content:
            # other possibilities: ImageFileContentBlock, ImageURLContentBlock
            if isinstance(content, TextContentBlock):
                self.response_text += content.text.value

    @override
    def on_tool_call_created(self, tool_call: ToolCall):
        if self.tool_call_completed(tool_call):
            return

        if self.auto_print:
            print(flush=True)
            print(HTML(
                f"Tool call created: <b><ansiblue>{tool_call.type}</ansiblue> <i>[ID: {tool_call.id}]</i></b>"), flush=True)

    @override
    def on_tool_call_delta(self, delta: ToolCallDelta, snapshot: ToolCall) -> None:
        if self.tool_call_completed(snapshot):
            return

        if delta.type == "code_interpreter" and delta.code_interpreter:
            if self.auto_print:
                print(delta.code_interpreter.input or "", end="", flush=True)

    @override
    def on_tool_call_done(self, tool_call: ToolCall) -> None:
        if self.tool_call_completed(tool_call, True):
            return

        if not self.auto_print:
            return

        if tool_call.type == "code_interpreter":
            self.syntax_highlight_code_interpreter(tool_call)

        print(HTML(
            f"Tool call completed: <b><ansiblue>{tool_call.type}</ansiblue> <i>[ID: {tool_call.id}]</i></b>"), flush=True)

    def tool_call_completed(self, tool_call: ToolCall, completed: bool = False) -> bool:
        """
        Updates the state of a ToolCall stored in the EventHandler based on its id.

        Parameters:
            tool_call (ToolCall): The ToolCall object to update the state for.
            completed (bool): Indicates whether the tool call has been completed upon this call. Default is False.

        Returns:
            completed (bool): Indicates whether or not the tool call has already been completed previously.

        Invoke on the following funcs:
        - on_tool_call_created: set_tool_call(tool_call)
        - on_tool_call_delta: set_tool_call(snapshot)
        - on_tool_call_done: set_tool_call(tool_call, True)
        """
        if tool_call.id in self._completed_tool_calls:
            # already maked as completed, return early
            return True

        # update object based on id
        self._tool_calls[tool_call.id] = tool_call

        # conditionally mark as completed
        if completed:
            self._completed_tool_calls.add(tool_call.id)

        # NOTE(justin): this func only returns True if the tool_call was marked as completed *prior* to the current invokation
        return False

    def syntax_highlight_code_interpreter(self, tool_call: ToolCall) -> None:
        assert tool_call.type == "code_interpreter"
        settings = Settings()  # type: ignore
        code_str = tool_call.code_interpreter.input
        if not code_str.endswith("\n"):
            print(flush=True)
        code_lines = len(code_str.splitlines())
        utils.delete_lines(code_lines)
        formatter = Terminal256Formatter(style=settings.code_style)
        highlighted_code = utils.highlight_code(code_str, formatter=formatter)
        print(ANSI(highlighted_code), end="", flush=True)


def init(api_key: str, verify: bool = True):
    """
    Initializes the OpenAI client with the provided API key.

    Parameters:
        api_key (str): The API key to initialize the OpenAI API client with.
        verify (bool): Verify the API key by making a test request to the API. Default is True.

    Raises:
        ValueError: If the API key is not provided or is not a string.
        openai.AuthenticationError: If the API key is invalid or the verification request fails.
    """
    if not api_key or not isinstance(api_key, str):
        raise ValueError("API key must be a non-empty string.")

    global Client
    if "Client" in globals():
        if Client.api_key == api_key:
            return

    Client = OpenAI(api_key=api_key)
    if verify:
        Client.models.list()


class FileInfo(NamedTuple):
    hash: str
    file_id: str
    cached: bool = False


def create_file(file_path: Path) -> FileObject:
    with open(file_path, 'rb') as file:
        file_response = Client.files.create(file=file, purpose="assistants")
        return file_response


def get_file_infos(files: List[Path]) -> List[FileInfo]:
    file_infos: List[FileInfo] = []
    for file_path in files:

        hash = utils.calculate_hash(file_path)
        assert isinstance(hash, str), \
            "Calculated hash should be of type 'str'. Ensure the 'intdigest' parameter is set to false."

        file_id = FileCache.get_file_id_by_hash(hash)
        if file_id is None:
            # not cached, upload new file
            file = create_file(file_path)
            info = FileInfo(hash, file.id,)
            file_infos.append(info)
            FileCache.set_hash_file_id(hash, file.id)
        else:
            # use cached file id
            info = FileInfo(hash, file_id, True)
            file_infos.append(info)

    FileCache.save()
    return file_infos


def create_vector_store_from_file_ids(name: str, file_ids: List[str]) -> VectorStore:
    return Client.beta.vector_stores.create(name=name, file_ids=file_ids)


def create_vector_store(name: str, file_paths: List[Path]) -> Resources:
    print(f"Creating vector store with {len(file_paths)} file(s).", end=" ")
    file_infos = get_file_infos(file_paths)
    file_ids = [info.file_id for info in file_infos]
    file_contents = {fp: FileUtils.read_str(fp) for fp in file_paths}

    cached_count = sum(1 for info in file_infos if info.cached)
    print(f"[{cached_count} file(s) already cached]" if cached_count > 0 else "")

    vector_store = create_vector_store_from_file_ids(name, file_ids)
    return Resources([vector_store.id], file_ids, file_contents)


def create_assistant(
    model: str,
    name: str,
    instructions: str,
    file_search: bool = False,
    interpret_code: bool = False,
    respond_with_json: bool = False,
    description: Optional[str] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None
) -> Assistant:
    tools = []
    if file_search:
        tools.append({"type": "file_search"})
    if interpret_code:
        tools.append({"type": "code_interpreter"})
    response_format = {"type": "json_object"} if respond_with_json else "auto"
    assistant = Client.beta.assistants.create(
        model=model,
        name=name,
        instructions=instructions,
        tools=tools,
        description=description,
        metadata={
            "created_by": utils.package_name(),
            "version": str(utils.get_version()),
        },
        top_p=top_p,
        temperature=temperature,
        response_format=response_format,  # type: ignore
    )
    return assistant


def get_assistant(id: str) -> Assistant:
    return Client.beta.assistants.retrieve(id)


def get_thread(id: str) -> Thread:
    return Client.beta.threads.retrieve(id)


def create_thread(resources: Resources, file_search: bool = False, code_interpreter: bool = False, send_messages: bool = False) -> Thread:
    # https://platform.openai.com/docs/api-reference/threads/createThread#threads-createthread-tool_resources

    messages: List[TCPMessage] = []
    if send_messages:
        msg = TCPMessage(
            role="user",
            content=textwrap.dedent(f"""
            The following {len(resources.file_contents)} messages will contain a file path, and the contents of the file.
            This information will be provided in the following json schema:
            {{
                'path': "file path",
                'contents': "file contents"
            }}
            This information will be used throughout the duration of the conversation.
        """))
        messages.append(msg)

        for path, content in resources.file_contents.items():
            content_obj = {
                "path": str(path),
                "content": content
            }
            content = json.dumps(content_obj)
            if len(content) > 256000:
                # print(f"Skipping content of file {path.name}, as it exceeds the maximum length.")
                continue
            msg = TCPMessage(content=content, role="user")
            messages.append(msg)

    tool_resources = ToolResources(
        file_search={"vector_store_ids": resources.vector_store_ids},
        # code_interpreter = {"file_ids": file_ids}
    )

    if not file_search and "file_search" in tool_resources:
        del tool_resources["file_search"]
    if not code_interpreter and "code_interpreter" in tool_resources:
        del tool_resources["code_interpreter"]

    return Client.beta.threads.create(messages=messages, tool_resources=tool_resources)


def get_thread_response(thread_id: str, assistant_id: str, prompt: str, auto_print: bool = False) -> str:
    Client.beta.threads.messages.create(
        thread_id=thread_id, content=prompt, role="user")

    event_handler = EventHandler(auto_print=auto_print)
    with Client.beta.threads.runs.stream(
        thread_id=thread_id,
        assistant_id=assistant_id,
        event_handler=event_handler
    ) as stream:
        stream.until_done()

    return event_handler.response_text


def set_file_cache(file_cache: FileCacheObj):
    global FileCache
    FileCache = file_cache



from typing import List
from prompt_toolkit.document import Document
from prompt_toolkit.validation import Validator, ValidationError


class NumericChoiceValidator(Validator):

    def __init__(self, valid_choices: List[int]):
        self.valid_choices = valid_choices

    def validate(self, document: Document):
        try:
            value = int(document.text)
            if value not in self.valid_choices:
                raise ValidationError(
                    message="That number isn't a valid choice.",
                    cursor_position=len(document.text)
                )
        except ValueError:
            raise ValidationError(
                message="Input the number corresponding with your choice.",
                cursor_position=len(document.text)
            )


class ChoiceValidator(Validator):

    def __init__(
        self,
        choices: List[str],
        is_blacklist: bool = False,
        allow_empty: bool = False,
        case_sensitive: bool = True,
        invalid_message: str = "The choice you have input is invalid."
    ):
        self.choices = choices
        self.is_blacklist = is_blacklist
        self.case_sensitive = case_sensitive
        self.invalid_message = invalid_message

        if allow_empty:
            self.choices.remove(
                "") if is_blacklist else self.choices.append("")

        if not case_sensitive:
            self.choices = [c.lower() for c in self.choices]

    def validate(self, document: Document):
        # input as a string
        choice = document.text if self.case_sensitive else document.text.lower()
        valid = choice not in self.choices if self.is_blacklist else choice in self.choices
        if not valid:
            raise ValidationError(
                message=self.invalid_message,
                cursor_position=len(document.text)
            )



import json
import os
import copy
import warnings
from dataclasses import dataclass, fields, field
from typing import Set, Optional, Dict, List, Any, ClassVar, Tuple
from openai import AuthenticationError
from pathlib import Path
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from pygments.styles import get_all_styles
from summawise import utils, ai
from summawise.utils import Singleton, ChoiceValidator
from summawise.data import DataMode
from summawise.files import utils as FileUtils
from summawise.api_objects import *


@dataclass
class Settings(metaclass=Singleton):
    api_key: str
    # NOTE(justin): deprecated in version 0.3.0
    assistant_id: str = field(repr=False)
    model: str
    compression: bool
    data_mode: DataMode
    code_style: str
    assistants: AssistantList
    threads: ThreadList

    DEPRECATED_FIELDS: ClassVar[Set[str]] = {"assistant_id"}
    DEFAULT_MODEL: ClassVar[str] = DEFAULT_MODEL
    DEFAULT_COMPRESSION: ClassVar[bool] = True
    DEFAULT_CODE_STYLE: ClassVar[str] = "monokai"
    DEFAULT_DATA_MODE: ClassVar[DataMode] = DataMode.BIN

    # NOTE(justin): This class functions as a singleton. Example usage anywhere:
    # settings = Settings() # type: ignore (dismiss warnings related to required arguments)
    # print(settings.to_dict()) # contains info established in main()

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Tuple["Settings", bool]:
        """
        Create a Settings object from a dictionary representation.

        Parameters:
            data (Dict[str, Any]): A dictionary containing the settings data.

        Returns:
            Tuple["Settings", bool]: A tuple containing the Settings object created from the data and a boolean indicating whether the settings should be updated.
        """
        key_count = len(data)
        expected_key_count = len(fields(Settings)) - \
            len(Settings.DEPRECATED_FIELDS)
        save = key_count != expected_key_count

        # NOTE(justin): deprecated in version 0.3.0
        assistant_id = data.pop("assistant_id", None)
        assistants = data.pop("assistants", [])
        assistants = AssistantList.from_dict_list(assistants)

        threads = data.pop("threads", [])
        threads = ThreadList.from_dict_list(threads)

        settings = Settings(
            assistant_id=assistant_id,
            assistants=assistants,
            threads=threads,
            model=data.pop("model", Settings.DEFAULT_MODEL),
            compression=data.pop("compression", Settings.DEFAULT_COMPRESSION),
            code_style=data.pop("code_style", Settings.DEFAULT_CODE_STYLE),
            data_mode=DataMode(
                data.pop("data_mode", Settings.DEFAULT_DATA_MODE.value)),
            **data
        )

        # version 0.3.0 backwards compatability (adding multi-assistant support)
        if settings.assistant_id:
            ai.init(settings.api_key, verify=False)
            api_assistant = ai.get_assistant(assistant_id)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                assistant = Assistant()
            assistant.apply_api_obj(api_assistant)
            assert assistant.name and assistant.instructions
            assistants.append(assistant)
            save = True

        return settings, save

    def to_dict(self) -> Dict[str, Any]:
        data = utils.asdict_exclude(self, Settings.DEPRECATED_FIELDS)
        data["assistants"] = self.assistants.to_dict_list()
        data["threads"] = self.threads.to_dict_list()
        data["data_mode"] = self.data_mode.value
        return data

    @staticmethod
    def init() -> "Settings":
        file = Settings.file()
        if file.exists():
            s_str = FileUtils.read_str(file)
            s_dict = json.loads(s_str)
            settings, save = Settings.from_dict(s_dict)
        else:
            settings = Settings.prompt()
            save = True

        # initialize openai api
        # TODO(justin): key verification after settings init in main
        ai.init(settings.api_key, verify=False)

        # automatically create (or update) default assistants that are added/changed in future versions
        for da in DEFAULT_ASSISTANTS:
            try:
                ea, idx = settings.assistants.get(da.name)
                if not ea or ea != da:
                    # assistant doesn't exist or has changed
                    assistant = copy.copy(da)
                    api_assistant = ai.create_assistant(
                        **da.to_create_params())
                    assistant.apply_api_obj(api_assistant)
                    if ea and idx != -1:
                        # assistant exists (matching default name), but the instructions are different
                        # this means it was updated, so we delete it and re-create it
                        assistant.created_at = ea.created_at
                        settings.assistants[idx] = assistant
                    else:
                        settings.assistants.append(assistant)
                    save = True
            except Exception as ex:
                print(utils.ex_to_str(ex, append=da.name, include_traceback=True))

        if save:
            Settings.save()

        return settings

    @staticmethod
    def save() -> "Settings":
        settings = Settings()  # type: ignore
        s_str = json.dumps(settings.to_dict(), indent=4)
        FileUtils.write_str(Settings.file(), s_str)
        return settings

    @staticmethod
    def file() -> Path:
        return utils.get_summawise_dir() / "settings.json"

    @staticmethod
    def prompt() -> "Settings":
        api_key = os.getenv("OPENAI_API_KEY", "")
        api_key = validate_api_key(api_key)

        while api_key is None:
            api_key = prompt("Enter your OpenAI API key: ")
            api_key = validate_api_key(api_key)
            if not api_key:
                print("The API key you entered is invalid. Try again.")

        valid_models = ["gpt-4o", "gpt-4", "gpt-4-turbo", "gpt-3.5-turbo"]
        model_validator = ChoiceValidator(valid_models, allow_empty=True)
        model_completer = WordCompleter(valid_models)
        model = prompt(
            f"Enter the OpenAI model to use [Default: {Settings.DEFAULT_MODEL}]: ", completer=model_completer, validator=model_validator)
        if not len(model.strip()):
            model = Settings.DEFAULT_MODEL

        all_styles = list(get_all_styles())
        style_validator = ChoiceValidator(all_styles, allow_empty=True)
        style_completer = WordCompleter(all_styles)
        style = prompt(
            f"Enter the code syntax highlighting style to use [Default: {Settings.DEFAULT_CODE_STYLE}]: ", completer=style_completer, validator=style_validator)
        if not len(style.strip()):
            style = Settings.DEFAULT_CODE_STYLE

        assistants: List[Assistant] = []
        for da in DEFAULT_ASSISTANTS:
            assistant = copy.copy(da)
            api_assistant = ai.create_assistant(**da.to_create_params())
            assistant.apply_api_obj(api_assistant)
            assistants.append(assistant)

        # TODO(justin): maybe add ability to select data mode. possibly a cli option to re-configure settings too.
        # it's not too important, for now it'll default to binary and can be changed manually.

        return Settings(
            api_key=api_key,
            model=model,
            assistant_id="",  # NOTE(justin): deprecated in version 0.3.0
            assistants=AssistantList(assistants),
            threads=ThreadList(),
            compression=Settings.DEFAULT_COMPRESSION,
            code_style=style,
            data_mode=Settings.DEFAULT_DATA_MODE,
        )


def validate_api_key(api_key: str) -> Optional[str]:
    try:
        ai.init(api_key)
        return api_key
    except AuthenticationError:
        return None



import inspect
import hashlib
import xxhash
from enum import Enum
from typing import Any, List, Union, NamedTuple, cast
from types import ModuleType
from pathlib import Path
from summawise.errors import ValueTypeError


class DataMode(Enum):
    JSON = "json"
    BIN = "binary"

    def ext(self) -> str:
        extensions = {
            DataMode.JSON: "json",
            DataMode.BIN: "bin"
        }
        try:
            return extensions[self]
        except KeyError:
            raise ValueError(f"Unsupported DataMode: {self}")


class DataUnit:
    """
    Utility class for converting sizes in bytes to human-readable string representations with appropriate units.

    Attributes:
        B (int): Size multiplier for bytes.
        KB (int): Size multiplier for kilobytes.
        MB (int): Size multiplier for megabytes.
        GB (int): Size multiplier for gigabytes.
        TB (int): Size multiplier for terabytes.

    Methods:
        _get_units(): Retrieve a list of available unit names based on class attributes.
        bytes_to_str(sz_bytes: int) -> str: Convert a size in bytes to a human-readable string representation.
    """
    __excludes__ = ["units"]

    B: int = 1
    KB: int = 2 ** 10
    MB: int = 2 ** 20
    GB: int = 2 ** 30
    TB: int = 2 ** 40

    units: List[str] = []

    @staticmethod
    def _get_units() -> List[str]:
        # "B", "KB", "MB", "GB", "TB"
        if len(DataUnit.units):
            return DataUnit.units
        units = [
            (name, value) for name, value in inspect.getmembers(DataUnit)
            if not name.startswith("__")
            and name not in DataUnit.__excludes__
            and not callable(value)
        ]
        units = sorted(units, key=lambda x: x[1])
        DataUnit.units = [name for name, _ in units]
        return DataUnit.units

    @staticmethod
    def bytes_to_str(sz_bytes: int) -> str:
        """
        Convert a size in bytes to a human-readable string representation.

        Parameters:
            sz_bytes (int): The size in bytes to convert to a human-readable string.

        Returns:
            str: A human-readable string representation of the size, including units (e.g. "1.23 MB").
        """
        assert sz_bytes >= 0, "sz_bytes must be non-negative"
        size, uidx = sz_bytes, 0
        units = DataUnit.units or DataUnit._get_units()
        while size >= DataUnit.KB and uidx < len(units) - 1:
            size /= float(DataUnit.KB)
            uidx += 1
        return f"{size:.2f} {units[uidx]}"


class _HashAlg(NamedTuple):
    module: ModuleType
    function_name: str

    def init(self) -> Any:
        alg_init = getattr(self.module, self.function_name)
        return alg_init()


class HashAlg(Enum):
    """An enumeration class representing various hashing algorithms and offering their implementation."""
    SHA_256 = _HashAlg(hashlib, "sha256")
    SHA3_256 = _HashAlg(hashlib, "sha3_256")
    XXH_32 = _HashAlg(xxhash, "xxh32")
    XXH_64 = _HashAlg(xxhash, "xxh64")
    XXH_128 = _HashAlg(xxhash, "xxh128")
    XXH3_128 = _HashAlg(xxhash, "xxh3_128")
    XXH3_64 = _HashAlg(xxhash, "xxh3_64")

    def calculate(
        self,
        _input: Union[Path, str, bytes],
        intdigest: bool = False
    ) -> Union[str, int]:
        """
        Calculate the hash of the input using the specified algorithm.

        Parameters:
            _input (Union[Path, str, bytes]): The input data to calculate the hash for.
            intdigest (bool): Whether to return the hash as an integer or a string.

        Returns:
            Union[str, int]: The calculated hash, either as a string or an integer.

        Raises:
            ValueError: If the hash algorithm/object is not valid.
            ValueTypeError: If the input data type is not one of Path, str, or bytes.
        """
        try:
            alg = cast(_HashAlg, self.value)
            hash_obj = alg.init()
            assert HashAlg.hash_obj_valid(hash_obj)
        except (AssertionError, AttributeError) as ex:
            raise ValueError("Invalid hash object.") from ex

        if isinstance(_input, (bytes, str)):
            _input = _input.encode() if isinstance(_input, str) else _input
            hash_obj.update(_input)
        elif isinstance(_input, Path):
            with open(_input, "rb") as f:
                for chunk in iter(lambda: f.read(8 * DataUnit.KB), b""):
                    hash_obj.update(chunk)
        else:
            raise ValueTypeError(_input, (Path, str, bytes))

        return (
            hash_obj.hexdigest() if not intdigest else
            int.from_bytes(hash_obj.digest(), byteorder="big")
        )

    @staticmethod
    def hash_obj_valid(alg: Any) -> bool:
        """Check if the given algorithm object is a valid hash algorithm."""
        updater = getattr(alg, "update", None)
        digester = getattr(alg, "digest", None)
        hexdigester = getattr(alg, "hexdigest", None)
        return (
            updater is not None and callable(updater) and
            digester is not None and callable(digester) and
            hexdigester is not None and callable(hexdigester)
        )



import click
from prompt_toolkit import prompt
from summawise import ai, utils
from summawise.settings import Settings


@click.group()
def thread():
    """Commands related to managing threads."""
    pass


@thread.command()
def list():
    """List your available threads."""
    settings = Settings()  # type: ignore
    for idx, thread in enumerate(settings.threads, start=1):
        print(f"{idx}) {thread.name}")


@thread.command()
@click.argument("thread_id", type=str)
def delete(thread_id: str):
    """
    Delete a thread.\n
    Provided ID can be the thread name, id, or number from 'list'.
    """
    settings = Settings()  # type: ignore

    thread, idx = settings.threads.get(thread_id)
    if not thread:
        print("Failed to identify thread to delete.")
        return

    del settings.threads[idx]
    settings.save()
    print(f"Deleted thread successfully: {thread.name}")


@thread.command()
@click.argument("thread_id", type=str)
@click.pass_context
def restore(ctx: click.Context, thread_id: str):
    """Restore a saved thread from its latest state."""
    settings = Settings()  # type: ignore
    debug = ctx.obj.get("DEBUG", False)

    thread, _ = settings.threads.get(thread_id)
    if not thread:
        print("No thread found matching provided identifier.")
        return

    assistant_name, assistant_id = thread.assistant
    assistant, _ = settings.assistants.get(assistant_name)
    if assistant is not None:
        assistant_id = assistant.id

    # get a summary of where the conversation (aka "thread") left off
    try:
        ai.get_thread_response(
            thread.id, assistant_id, "Please summarize our conversation thus far.", auto_print=True)
    except Exception as ex:
        print(
            f"Error initializing conversation: {utils.ex_to_str(ex, include_traceback=debug)}")
        return

    # resume interactive prompt (from scan command) with restored thread
    print("\nYou can now continue to ask questions about the content. Type 'exit' to quit.")
    while True:
        input_str = prompt("\nyou > ")
        utils.conditional_exit(input_str)
        try:
            ai.get_thread_response(
                thread.id, assistant_id, input_str, auto_print=True)
        except Exception as ex:
            print(
                f"\nError occurred during conversation: {utils.ex_to_str(ex, include_traceback=debug)}")



from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterable, Tuple
from openai.types.beta import Thread as APIThread
from summawise import utils, ai
from summawise.api_objects.generic import ApiObjList, BaseApiObj


@dataclass
class Thread(BaseApiObj):
    id: str
    name: str
    assistant: Tuple[str, str]  # name, id
    created_at: datetime = field(default_factory=utils.utc_now)

    def get_api_obj(self) -> APIThread:
        if not self.id:
            raise ValueError(
                "Thread ID must be specified to retrieve API object.")
        api_thread = ai.get_thread(self.id)
        return api_thread


class ThreadList(ApiObjList[Thread]):

    def __init__(self, threads: Iterable[Thread] = [], **kwargs):
        super().__init__(threads, cls=Thread, **kwargs)

    @staticmethod
    def from_dict_list(  # type: ignore
        threads: Iterable[Thread],
        **kwargs
    ) -> "ThreadList":
        return ApiObjList.from_dict_list(
            objects=threads,  # type: ignore
            cls=Thread,
            **kwargs
        )



import warnings
from typing import Dict, List, Optional, TypeVar, NamedTuple, Iterable
from datetime import datetime, timezone
from dataclasses import dataclass, field
from openai.types.beta import Assistant as APIAssistant
from summawise import utils
from summawise.data import HashAlg
from summawise.api_objects.generic import ApiObjList, BaseApiObj

T = TypeVar('T')

DEFAULT_MODEL = "gpt-3.5-turbo"


@dataclass
class Assistant(BaseApiObj):
    id: str = ""
    name: str = ""
    instructions: str = ""
    model: str = DEFAULT_MODEL
    file_search: bool = False
    interpret_code: bool = False
    respond_with_json: bool = False
    description: Optional[str] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    created_at: datetime = field(default_factory=utils.utc_now)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Assistant):
            return NotImplemented
        return (
            self.name == other.name and
            self.instructions == other.instructions
        )

    def __hash__(self) -> int:
        strval = f"{self.name}\n{self.instructions}"
        hashval = HashAlg.XXH_64.calculate(strval, intdigest=True)
        return int(hashval)

    def __post_init__(self):
        def missing_field_warning(x): return warnings.warn(
            f"'Assistant' object is missing expected field {x}.\nThis field should always be provided.", UserWarning)
        if not self.name:
            missing_field_warning("name")
        if not self.instructions:
            missing_field_warning("instructions")

    def to_create_params(self) -> dict:
        return utils.asdict_exclude(self, {"id", "created_at"})

    def apply_api_obj(self, obj: APIAssistant):
        """Use official library API object to apply/overwrite specific fields to this dataclass."""
        self.id = obj.id
        self.name = obj.name or ""
        self.instructions = obj.instructions or ""
        self.model = obj.model
        self.description = obj.description
        self.top_p = obj.top_p
        self.temperature = obj.temperature
        self.created_at = datetime.fromtimestamp(obj.created_at, timezone.utc)


class AssistantList(ApiObjList[Assistant]):

    def __init__(self, assistants: Iterable[Assistant] = [], **kwargs):
        super().__init__(assistants, cls=Assistant, **kwargs)

    @staticmethod
    def from_dict_list(  # type: ignore
        assistants: List[dict],
        **kwargs
    ) -> "AssistantList":
        return ApiObjList.from_dict_list(  # type: ignore
            objects=assistants,
            cls=Assistant,
            **kwargs
        )


class ConversationInit(NamedTuple):
    user_msg: str
    gpt_msg: str


TranscriptAnalyzer: Assistant = Assistant(
    name="Transcript Analysis Assistant",
    instructions="You are an assistant that summarizes video transcripts and answers questions about them.",
    file_search=True
)

CodingAssistant: Assistant = Assistant(
    name="Coding Assistant",
    instructions=(
        "You are an assistant which helps users by generating, debugging, and optimizing code snippets.\n"
        "It provides clear, well-commented code along with explanations to aid user understanding.\n"
        "\nGeneral Guidelines\n"
        "1.) Understand the Request: Ensure you fully understand the users requirements. Ask clarifying questions if needed.\n"
        "2.) Simplicity and Clarity: Provide simple, clean, and well-documented code. Avoid overly complex solutions unless explicitly requested.\n"
        "3.) Explain the Code: Always include a brief explanation of the code to help the user understand how it works.\n"
        "4.) Best Practices: Follow coding best practices and conventions for the relevant programming language.\n"
        "5.) Error Handling: Include error handling where appropriate to make the code robust.\n"
        "6.) Modularity: Write modular code with functions or classes to enhance readability and reusability.\n"
        "\nCode Generation Steps\n"
        "1.) Identify the Language: Confirm the programming language to be used (e.g., Python, JavaScript, etc.).\n"
        "2.) Gather Requirements: Understand the specific functionality or problem the user needs the code to address.\n"
        "3.) Write the Code: Generate the code snippet or full program.\n"
        "4.) Comment the Code: Add comments to explain key parts of the code.\n"
        "5.) Provide Explanation: Write a brief explanation of what the code does and how it works.\n"
        "6.) Optimize: If applicable, suggest optimizations or improvements."
    ),
    file_search=True,
    interpret_code=True
)

CONVERSATION_INITS: Dict[Assistant, ConversationInit] = {
    TranscriptAnalyzer: ConversationInit(
        "Generating summary of transcript...",
        "Please summarize the transcript."
    ),
    CodingAssistant: ConversationInit(
        "Generating summary of codebase...",
        "Please summarize the codebase."
    )
}

DEFAULT_ASSISTANTS: List[Assistant] = list(CONVERSATION_INITS.keys())
